/*****************************************************************************

 JOCO's ZX spectrum emulator.
 zx_hw.c: the spectrum hardware layer implementation, 
 I/O ports (ULA), and screen realization

 see the file read.me for details/legal

******************************************************************************/

#include <stdio.h>
#include <dpmi.h>
#include <go32.h>
#include <pc.h>

#include "keytab.h"
#include "z80cpu.h"

typedef unsigned int  RGBval;

word DescA000;
int  SAddrTab[6144];
int  AAddrTab[ 768];
byte FlashCnt=20;
byte BorderColor;
byte PCKBD[256];

word ScrAddr1=0x0000;
word ScrAddr2=0x1f40*2;

/* You need to understand this: We use the 80 chars wide texmode where
   the logical scanline widht is 640 pixels. So each spectrum pixel 
   is composed by two PC pixels. For one spectrum screenbyte we need
   2 PC chars where the charheight is 2 scanlines! (beacuse the 80x25
   textmode screen is 400 scanlines height. That's why each value is
   duplicated.)
   This table contains the pixels for spectrum nibbles 0x00..0x0F. 
   Each spectrum screenbyte is placed in two PC bytes. 
   (see how I use this table in the SpectrumScreen() function in order 
    to set the used chars)
*/
byte gfbits[]={	0x00,0x00,0x03,0x03,0x0c,0x0c,0x0f,0x0f,
		0x30,0x30,0x33,0x33,0x3c,0x3c,0x3f,0x3f,
		0xc0,0xc0,0xc3,0xc3,0xcc,0xcc,0xcf,0xcf,
		0xf0,0xf0,0xf3,0xf3,0xfc,0xfc,0xff,0xff};

/* here we'll store the spectrum screen bytes. See Sput() below */
int gfxb[256];

/* These are the VGA RGB values for the Spectrum's colors.
   Fortunately the Spectum has only 15 and (not 16 colors), cause the
   brighted black is same as the normal black. In this way I can use the
   0x00 color for the Spectrum's border */
   byte RGB_palette[48]={
       00,00,00,      // 0. BORDER color!      
       00,00,35,      // 1. ZX Blue            
       40,00,00,      // 2. ZX Red             
       40,00,40,      // 3. ZX Magenta         
       00,40,00,      // 4. ZX Green           
       00,40,40,      // 5. ZX Cyan            
       50,50,00,      // 6. ZX Yellow          
       50,50,50,      // 7. ZX White           
       00,00,00,      // 8. ZX Black           
       00,00,50,      // 9. ZX BRIGHT Blue     
       63,00,00,      //10. ZX BRIGHT Red      
       63,00,63,      //11. ZX BRIGHT Magenta  
       00,63,00,      //12. ZX BRIGHT Green    
       00,63,63,      //13. ZX BRIGHT Cyan     
       63,63,00,      //14. ZX BRIGHT Yellow   
       63,63,63};     //15. ZX BRIGHT White    

/* We use this table for convert a Spectrum color to a VGA color */
   byte ColorConvTab[256]={
       // NORMAL colors 
       0x88,0x81,0x82,0x83,0x84,0x85,0x86,0x87,
       0x18,0x11,0x12,0x13,0x14,0x15,0x16,0x17,
       0x28,0x21,0x22,0x23,0x24,0x25,0x26,0x27,
       0x38,0x31,0x32,0x33,0x34,0x35,0x36,0x37,
       0x48,0x41,0x42,0x43,0x44,0x45,0x46,0x47,
       0x58,0x51,0x52,0x53,0x54,0x55,0x56,0x57,
       0x68,0x61,0x62,0x63,0x64,0x65,0x66,0x67,
       0x78,0x71,0x72,0x73,0x74,0x75,0x76,0x77,
       // BRIGHTed colors 
       0x88,0x89,0x8a,0x8b,0x8c,0x8d,0x8e,0x8f,
       0x98,0x99,0x9a,0x9b,0x9c,0x9d,0x9e,0x9f,
       0xa8,0xa9,0xaa,0xab,0xac,0xad,0xae,0xaf,
       0xb8,0xb9,0xba,0xbb,0xbc,0xbd,0xbe,0xbf,
       0xc8,0xc9,0xca,0xcb,0xcc,0xcd,0xce,0xcf,
       0xd8,0xd9,0xda,0xdb,0xdc,0xdd,0xde,0xdf,
       0xe8,0xe9,0xea,0xeb,0xec,0xed,0xee,0xef,
       0xf8,0xf9,0xfa,0xfb,0xfc,0xfd,0xfe,0xff,
       // NORMAL colors (with flash)
       0x88,0x81,0x82,0x83,0x84,0x85,0x86,0x87,
       0x18,0x11,0x12,0x13,0x14,0x15,0x16,0x17,
       0x28,0x21,0x22,0x23,0x24,0x25,0x26,0x27,
       0x38,0x31,0x32,0x33,0x34,0x35,0x36,0x37,
       0x48,0x41,0x42,0x43,0x44,0x45,0x46,0x47,
       0x58,0x51,0x52,0x53,0x54,0x55,0x56,0x57,
       0x68,0x61,0x62,0x63,0x64,0x65,0x66,0x67,
       0x78,0x71,0x72,0x73,0x74,0x75,0x76,0x77,
       // BRIGHTed colors (with flash)
       0x88,0x89,0x8a,0x8b,0x8c,0x8d,0x8e,0x8f,
       0x98,0x99,0x9a,0x9b,0x9c,0x9d,0x9e,0x9f,
       0xa8,0xa9,0xaa,0xab,0xac,0xad,0xae,0xaf,
       0xb8,0xb9,0xba,0xbb,0xbc,0xbd,0xbe,0xbf,
       0xc8,0xc9,0xca,0xcb,0xcc,0xcd,0xce,0xcf,
       0xd8,0xd9,0xda,0xdb,0xdc,0xdd,0xde,0xdf,
       0xe8,0xe9,0xea,0xeb,0xec,0xed,0xee,0xef,
       0xf8,0xf9,0xfa,0xfb,0xfc,0xfd,0xfe,0xff};

/* Calling BIOS function INT 10h to set a videomode */
void vga_mode(int mode_nr) {
 __dpmi_regs 	r;
  
 r.d.eax=mode_nr&0x7f;
 __dpmi_int(0x10,&r);
}


/* Calling BIOS INT 10h to turn off the PC blinking, so we can 
   use 16 fore and 16 back colors in textmodes */
void blinking_bit(int v) {
 __dpmi_regs 	r;
  
 r.d.eax=0x1003;
 r.d.ebx=v&1;
 __dpmi_int(0x10,&r);
}

/* a simple copy function which copies bytes to the DOS memory (lower 1M) */
void copy_to_dosmem(word selector, int offset, void *source, int size) {
 __asm __volatile__("
	push	%%ES
	pushal
	movw	%0,%%ax
	movw	%%ax,%%ES
	movl	%1,%%edi
	movl	%2,%%esi
	movl	%3,%%ecx
	cld
	rep
	movsb
	popal
	pop	%%ES
 "
 :
 : "g" (selector), "g" (offset), "g" (source), "g" (size)
 : "%eax");
}

/* Change the character set */
void set_chars(void *chrdata, int chrnum, int chrheight) {
 __dpmi_regs 		r;
 _go32_dpmi_seginfo     s;
 int i;

 r.x.ax=0x1110;			/* INT 10h BIOS function */
 r.x.bx=(chrheight&0xff)<<8;
 r.x.cx=chrnum;
 r.x.dx=0x0000;

 memset(&s,0x00,sizeof(s));
 s.size=chrnum*chrheight;
 i=_go32_dpmi_allocate_dos_memory(&s);
 copy_to_dosmem(s.pm_selector,s.pm_offset,chrdata,chrnum*chrheight);

 r.x.bp=s.rm_offset;
 r.x.es=s.rm_segment;

 __dpmi_int(0x10,&r);

 _go32_dpmi_free_dos_memory(&s);
}

/* turn off the 9th bit of the characters, as we need 8px wide chars */
void box8() {
 int i;

 outportb(0x3c2,inportb(0x3cc)&0xf3);
 outportb(0x3c4,0x01);
 outportb(0x3c5,inportb(0x3c5)|0x01);

}

/* Assign a VGA palette register to a legacy EGA palette register */
void setEGApal(int reg, int data) {
 int i;

 i=inportb(0x3da);
 outportb(0x3c0,reg&0x0f);
 outportb(0x3c0,data&0xff);
 outportb(0x3c0,0x20);
}

/* Define a VGA color with R-G-B components */
void setVGApal(int reg,int R,int G,int B) {
 outportb(0x3c8,reg);
 outportb(0x3c9,R);
 outportb(0x3c9,G);
 outportb(0x3c9,B);
}

/* Clear the video memory 0xA0000(64K) */
void ClearVideoArea() {
 __asm __volatile__("
	push	%ES
	pushal
	movw	_DescA000,%ES
	cld
	xorl	%edi,%edi
	xorl	%eax,%eax
	movl	$16384,%ecx	/* 4*16K = 64K ! */
	rep	
	stosl
	popal
	pop	%ES
 ");	
}

/* A high-speed routine to show one spectrum screen byte on the 
   PC screen. Called by the CPU emulator (SpecWriteByte).
   Some words about the flash emulation: I use two screens and change them
   by the timer interrupt. Where need to flash, the colors are exchanged on
   the secondary screen. Basic idea, isn't it. Ok, but because of this,
   I have to put all the spectrum screenbytes to both of the screens. */
void Sput(int Addr, int Data) {
 __asm __volatile__("
	push	%%ES
	movw	_DescA000,%%ES
	movl	%0,%%ebx

	subl	$16384,%%ebx			/* %%ebx: absoulte address */
	movl	_SAddrTab(,%%ebx,4),%%ebx	/* real PC screen position */

	movl	%1,%%eax
	movl	_gfxb(,%%eax,4),%%eax		/* get the wide pixels     */

	movb	%%ah,%%ES:(%%ebx)		/* primary screen nybbles  */
	movb	%%al,%%ES:2(%%ebx)

	movb	%%ah,%%ES:32000(%%ebx)		/* secondary screen nybbles*/
	movb	%%al,%%ES:32002(%%ebx)
	pop	%%ES
 "
 :
 : "g" (Addr), "g" (Data)
 : "%eax","%ebx");
}

/* A high-speed routine to show one spectrum attr. byte on the 
   PC screen. Called by the CPU emulator (SpecWriteByte). In this case
   we have to modify 16 PC attribute bytes on both screens */
void Aput(int Addr, int Color) {
 __asm __volatile__("
	push	%%ES
	movw	_DescA000,%%ES

	movl	%1,%%ebx			/* get the Color 	*/
	movb	_ColorConvTab(,%%ebx,1),%%al	/* convert it !  	*/
	movb	%%al,%%ah			
	test	$128,%%bl                       /* flash on??    	*/
	jz	no_flash
	ror	$4,%%ah                         /* xchg nybbles if flash   */
no_flash:
	movl	%0,%%ebx			/* get the Address 	   */
	subl	$22528,%%ebx			/* absolute address        */
	movl	_AAddrTab(,%%ebx,4),%%ebx	/* real PC screenpos       */

	movb	%%al,%%ES:(%%ebx)		/* primary screen update   */
	movb	%%al,%%ES:2(%%ebx)
	movb	%%al,%%ES:160(%%ebx)
	movb	%%al,%%ES:162(%%ebx)
	movb	%%al,%%ES:320(%%ebx)
	movb	%%al,%%ES:322(%%ebx)
	movb	%%al,%%ES:480(%%ebx)
	movb	%%al,%%ES:482(%%ebx)
	movb	%%al,%%ES:640(%%ebx)
	movb	%%al,%%ES:642(%%ebx)
	movb	%%al,%%ES:800(%%ebx)
	movb	%%al,%%ES:802(%%ebx)
	movb	%%al,%%ES:960(%%ebx)
	movb	%%al,%%ES:962(%%ebx)
	movb	%%al,%%ES:1120(%%ebx)
	movb	%%al,%%ES:1122(%%ebx)

	movb	%%ah,%%ES:32000(%%ebx)		/* secondary screen update */
	movb	%%ah,%%ES:32002(%%ebx)
	movb	%%ah,%%ES:32160(%%ebx)
	movb	%%ah,%%ES:32162(%%ebx)
	movb	%%ah,%%ES:32320(%%ebx)
	movb	%%ah,%%ES:32322(%%ebx)
	movb	%%ah,%%ES:32480(%%ebx)
	movb	%%ah,%%ES:32482(%%ebx)
	movb	%%ah,%%ES:32640(%%ebx)
	movb	%%ah,%%ES:32642(%%ebx)
	movb	%%ah,%%ES:32800(%%ebx)
	movb	%%ah,%%ES:32802(%%ebx)
	movb	%%ah,%%ES:32960(%%ebx)
	movb	%%ah,%%ES:32962(%%ebx)
	movb	%%ah,%%ES:33120(%%ebx)
	movb	%%ah,%%ES:33122(%%ebx)

	pop	%%ES
 "
 :
 : "g" (Addr), "g" (Color)
 : "%eax","%ebx");
}

/* Flash emulation: this function is called by the timer ISR. It exchnages
   the primary and secondary screens, so what you see is flashing. */
void Emulate_Flash() {
 __asm __volatile__("
   decb   _FlashCnt
   jnz    quit
   movb   $20,_FlashCnt
   movw   _ScrAddr1,%%bx	/* xchg the addresses */
   xchg   %%bx,_ScrAddr2
   movw   %%bx,_ScrAddr1
   movw   $0x03D4,%%dx  	/* update the VGA screen start address */
   movb   $0x0c,%%al		/* registers */
   movb   %%bh,%%ah
   outw   %%ax,%%dx
   incb   %%al
   movb   %%bl,%%ah
   outw   %%ax,%%dx
quit:
"
 :
 : 
 : "%eax","%edx");
}

/* in case we'd need to refresh the screen "manually" */
void RefreshScreen() {
 int i;

 for(i=16384;i<22528;i++) Sput(i,Z80mem[i]);
 for(i=22528;i<22528+768;i++) Aput(i,Z80mem[i]);

 //while (!PCKBD[1]);
 //while (PCKBD[1]);
}

/*
  Called by the CPU emulator. For now it's simple: we just get the
  pressed key bits from the ZXKBD array and ANDing them.
*/
// ecx: port address, al: data return
void ULAin() {
 __asm __volatile__("
	pushl	%ecx
	pushl	%edi
	movb	$0xbf,%al		/* default input value for ULA    */
	movb	$8,%cl			/* speccy has 8 keyboard rows     */
	leal	_ZXKBD,%edi             /* source bitarray                */
in_loop:
	shrb	$1,%ch			/* the the keyrow if the selector */
	jc	noread			/* bit is 0 			  */
	andb	(%edi),%al		/* making the result byte 	  */
noread:
	incl	%edi
	decb	%cl
	jnz	in_loop
	popl	%edi
	popl	%ecx
 ");
}

/*  Called by the CPU emualtor. We handle the speaker bit, and the border
    colors. */ 
// ecx: port address, al: data to write to 
void ULAout() {
 __asm __volatile__("
	pushal
	movb	%al,%cl
	inb	$0x61,%al
	testb	$0x10,%cl		/* ULA speaker bit */
	jz	spk_off
spk_on:
	orb	$3,%al
	outb	%al,$0x61
	jmp	setborder
spk_off:
	and	$0xfc,%al
	outb	%al,$0x61
setborder:
	and	$7,%ecx			/* ULA border bits */
	cmpb	_BorderColor,%cl	/* same? */
	jz	ulaout_quit
	movb    %cl,_BorderColor
	mov	$0x3c8,%edx     	/* VGA palette write */
	xorb	%al,%al
	outb	%al,%dx			/* 0x00:  our border */
	incl	%edx
					/* tricky: ecx*3!!   */
	movl	_RGB_palette(%ecx,%ecx,2),%ecx	/* ecx=RGB   */
	mov	%cl,%al
	outb	%al,%dx			/* Reg 		*/
	mov	%ch,%al
	outb	%al,%dx			/* Green 	*/
	shrl	$16,%ecx
	mov	%cl,%al
	outb	%al,%dx			/* Blue 	*/
ulaout_quit:
	popal
 ");
}

/* Timer ISR function. Emulates the flash effect and sets the INT bit in
   the CPU emulator's sysflag. It's a 50IRQs/sec period here! */
void my_timer_isr(void) {
 Emulate_Flash();
 HWFlag|=0x01;
}

/* Timer ISR installer/deinstaller: see DJGPP docs for more info.
   Protected mode sucks. */

 _go32_dpmi_seginfo OldINT_1C,NewINT_1C;

void InstallTimerISR() {
  outportb(0x43,0xb6);		/* timer reset */
  outportb(0x43,0);
  outportb(0x43,0);

  outportb(0x43,0x36);		/* timer set to 50irqs/sec */
  outportb(0x40,23684&255);
  outportb(0x40,23684>>8);
  /* protected mode stuff comes */
  NewINT_1C.pm_offset =(int)my_timer_isr;
  NewINT_1C.pm_selector=_go32_my_cs();
  _go32_dpmi_get_protected_mode_interrupt_vector(0x1c,&OldINT_1C);
  _go32_dpmi_allocate_iret_wrapper(&NewINT_1C);
  _go32_dpmi_set_protected_mode_interrupt_vector(0x1c,&NewINT_1C);
}

void UninstallTimerISR() {
  outportb(0x43,0x36);		/* timer reset */
  outportb(0x40,0x00);
  outportb(0x40,0x00);
  /* protected mode stuff comes */
  _go32_dpmi_set_protected_mode_interrupt_vector(0x1c,&OldINT_1C);
  _go32_dpmi_free_iret_wrapper(&NewINT_1C);
}

/* Calculate the screen positions for the spectrum screenbytes.
   Don't try to understand this. I don't get it either. :-) */
void CalcScrnAddr() {
int i,j,k,l,m,w;

 m=0;
 for(i=0;i<3;i++) 
  for(j=0;j<8;j++) 
   for(k=0;k<8;k++) 
    for(l=0;l<32;l++) {
     w=4*160+8*2+ 4*(l+k*8*40+j*40+i*64*40);
     SAddrTab[m]=w; m++;
    }
}

/* Calculate the attr addresses for the spectrum attr bytes. */
void CalcAttrAddr() {
int i,j,l,m,w;

 m=0;
 for(i=0;i<3;i++) 
  for(j=0;j<8;j++)
   for(l=0;l<32;l++) {
    w=4*160+8*2+1+ 4*(l+j*8*40+i*64*40);
    AAddrTab[m]=w; m++;
   }
}

/* 
  Change the sructure of the 80x25 textmode screen:
  this is the point where we start using non-standard video mode.
  In this videomode the charheight is 2 scanlines, so virtually we
  have an 80x200 charscreen. The normal 80x25 textmode uses 720 pixels
  horizontally (chars are 9px wide), but we need 640px wide screen.
  We're gonna use this format to produce the ZX spectrum screen.
*/
void SpectrumScreen() {
 int i;

 vga_mode(0x03);	/* start: 80x25 textmode */
 blinking_bit(0);	/* no blinking, 16 fg/bg on */
 set_chars(gfbits,16,2);/* set the first 16 chars, spectrum patterns */
 /* make the 256 possible screenbyte combinations */
 for(i=0;i<256;i++) gfxb[i]=((i>>4)<<8)|(i&0x0f);
 /* 720px wide mode off, 640px wide mode on ! */
 box8();
 /* palette init. piece of cake */
 for(i=0;i<16;i++) {
  setEGApal(i,i);
  setVGApal(i,RGB_palette[i*3],RGB_palette[i*3+1],RGB_palette[i*3+2]);
 }
 /* here comes a "magic" step. Normal textmodes use the 0xB800 segment,
    but this one only 32K long. Not enough! We need 2x2x80x200 bytes.
    So we're changing the screen position to 0xA0000 and the size is
    64K */
 outportw(0x3ce,0x0606);
 /* fill the conversion tables */
 CalcScrnAddr();
 CalcAttrAddr();
 /* clear the video memory */
 ClearVideoArea();
}

/* Restore the standard 80x25 textmode screen */
void PcScreen() {
 vga_mode(0x03);
 blinking_bit(0);
}

/* Keyboard ISR. We got the keyboard state from the hardware, and here
   we update the ZXKBD table */
void my_keyboard_isr() {

 unsigned char b,c,d;
 b=inportb(0x60);		/* scan-code */
 PCKBD[b&0x7f]=(b>>7)^1;	/* press/release state */

 c=Key_tab[(b&127)-1];		/* convert it */
 if (c&0x80) // Caps Shift press/release
  (b&0x80)?(ZXKBD[0]|=0x01):(ZXKBD[0]&=0xfe);
 if (c&0x40) // Symbol Shift press/release
  (b&0x80)?(ZXKBD[7]|=0x02):(ZXKBD[7]&=0xfd);

 if (c!=99) { // if it's a valid spectrum key...
  c&=7;
  d=Key_tab_bit[(b&127)-1];	/* update the ZXKBD table */
  (b&0x80)?(ZXKBD[c]|=d):(ZXKBD[c]&=~d);
 }

 // end of IRQ, hardware stuff
 b=inportb(0x61);
 outportb(0x61,b|0x80);
 outportb(0x61,b);
 outportb(0x20,0x20);
}


/* Keyboard ISR installer/deinstaller: see DJGPP docs for more info.
   Protected mode sucks. */

_go32_dpmi_seginfo OldINT_09,NewINT_09;

void InstallKeyboardISR() {
  NewINT_09.pm_offset =(int)my_keyboard_isr;
  NewINT_09.pm_selector=_go32_my_cs();
  _go32_dpmi_get_protected_mode_interrupt_vector(0x09,&OldINT_09);
  _go32_dpmi_allocate_iret_wrapper(&NewINT_09);
  _go32_dpmi_set_protected_mode_interrupt_vector(0x09,&NewINT_09);
}

void UninstallKeyboardISR() {
  _go32_dpmi_set_protected_mode_interrupt_vector(0x09,&OldINT_09);
  _go32_dpmi_free_iret_wrapper(&NewINT_09);
}
